/*-
 * Copyright (c) 2005 Poul-Henning Kamp
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: phk_icsp.c,v 1.2 2010/02/27 13:00:25 phk Exp $
 *
 */

#define SERIAL 0
#ifdef FLEXELINT
typedef unsigned char __sfr;
#endif

#include "pic18fregs.h"

#include <stdint.h>


#ifndef CTASSERT                /* Allow lint to override */
#define CTASSERT(x)             _CTASSERT(x, __LINE__)
#define _CTASSERT(x, y)         __CTASSERT(x, y)
#define __CTASSERT(x, y)        typedef char __assert ## y[(x) ? 1 : -1]
#endif

#pragma stack 0x200 0x100

/* Enable the watchdog at 4msec * 64 = .256 sec */
code unsigned char at(__CONFIG2H) config2h = (0x7 << 1) | 1;

/* HSPLL */
code unsigned char at(__CONFIG2L) config2l = 0
	| (1 << 7)	// IESO two speed startup
	| (1 << 6)	// FCMEM
	| (1 << 4)	// LPT1OSC
	| (1 << 3)	// T1DIG
	| (3 << 0)	// ECPLL CLKO@RA6
	;

/* No code protects.  No CPU system clock divde */
code unsigned char at(__CONFIG1H) config1h = 0
	| (0xf0)	// unimplemented
	| (1 << 2)	// CP0
	| (3 << 0)	// CPDIV = /1
	;

/* No debugger, no extins, reset on stack, WDT enable, PLLDIV=? */
code unsigned char at(__CONFIG1L) config1l = 0
	| (1 << 7)	// #DEBUG
	| (0 << 6)	// XINST
	| (1 << 5)	// STVREN
	| (6 << 1)	// PLLDIV = 8 MHz
	| (1 << 0)	// WDTEN
	;

#define MHZ	48

/* Serial port defines -----------------------------------------------*/

#if SERIAL
#define SERIALPORTS (1 << 1)

#include "serial.c"

#endif

#include "usb.c"

/*********************************************************************/
static char txBuffer[64];
static uint8_t txbp;
static char rxBuffer[64];
static uint8_t rxbp, rxbe;

static uint16_t rate;

// Regardless of what the USB is up to, we check the USART to see
// if there's something we should be doing.
static void
USBEcho(void)
{
	uint8_t c, j;

	if ((deviceState < CONFIGURED) || (UCONbits.SUSPND==1))
		return; 

	if (rxbp == rxbe) {
		rxbe = OutPipe(1, rxBuffer, sizeof rxBuffer);
		rxbp = 0;
	}

	while (rxbp < rxbe) {
		c = rxBuffer[rxbp++];
		if ((c & 0xf0) == 0x40) {
			for (j = 0; j < rate; j++)
				LATB = (c & 0x0f) << 2;
		} else if (c == 'r') {
			TRISBbits.TRISB2 = 1;		// PGD
		} else if (c == 'w') {
			TRISBbits.TRISB2 = 0;		// PGD
		} else if (c == '!') {
			txBuffer[txbp++] = c;
		} else if (c == '?') {
			for (j = 0; j < rate; j++)
				c = PORTB;
			c >>= 2;
			c &= 0xf;
			c |= 0x60;
			txBuffer[txbp++] = c;
		}
	} 

	if (txbp > 0) {
		j = InPipe(1, txBuffer, txbp);
		if (j == 0) {
		} else if (j == txbp) {
			txbp = 0;
		} else {
			txbp = 0;
		}
	}
}

/*********************************************************************/

void
intr_h() interrupt 1
{
	uint8_t u;

	u = PIR2;
	if (u & 0x10)
		USB_intr();
	PIR2 = 0;
}

void
intr_l() interrupt 2
{

}

/*********************************************************************/

void
main(void) wparam
{
	uint16_t u;

#if SERIAL
	TRISCbits.TRISC6 = 0;
	SERIAL_INIT(1);
	SERIAL_BAUD(1,103);
	TXSTA1bits.BRGH = 1;
	BAUDCON1bits.BRG16 = 1;
	stdin = STREAM_USART;
	stdout = STREAM_USART;
	debugLevel = 127;
#else

	stdin = STREAM_USER;
	stdout = STREAM_USER;
	debugLevel = 0;
#endif
	OSCTUNEbits.PLLEN = 1;

	/* Wait for PLL to lock */
	for (u = 0; u < 0x8000; u++)
		;

	// Set all I/O pins to digital
	ANCON1 |= 0x1F;

	LATB = 0x8 << 2;
	TRISBbits.TRISB5 = 0;		// RST
	TRISBbits.TRISB4 = 0;		// PGM
	TRISBbits.TRISB3 = 0;		// PGC
	TRISBbits.TRISB2 = 1;		// PGD

	// Initialize USB
	UCFG = 0x14; // Enable pullup resistors; full speed mode
	deviceState = DETACHED;
	remoteWakeup = 0x00;
	currentConfiguration = 0x00;

	INTCON2bits.RBPU = 0;		// Weak pull-up

	/* Setup Interrupts */
	RCONbits.IPEN = 1;
	INTCON = 0xc0;
	PIE2bits.USBIE = 1;

	rate = 20;
	txbp = 0;
	rxbp = 0;
	rxbe = 0;

	while(1) {
		ClrWdt();
		EnableUSBModule();
		USBEcho();
	}
}
